/*
 * MZViewer - A viewer for MZData format Mass Spectra
 * Copyright (c) 2005 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
package uk.ac.bbsrc.babraham.mzdata.displays;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.text.DecimalFormat;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import uk.ac.bbsrc.babraham.mzdata.*;
import uk.ac.bbsrc.babraham.mzdata.exceptions.MZSpectrumException;

/**
 * @author andrewss
 *
 */
public class ChromatogramDisplayPanel extends JPanel {
    
    private MZCollection collection;
    private MZSpectrum [] spectra;
    private int [] msLevels;
    private int currentMSlevel;
    private double maxIntensity = 0;
    private double minIntensity = 0;
    private float maxTime = 0;
    private float minTime = 0;
    private boolean changingXZoom = false;
    private boolean changingYZoom = false;
    private int dragXstarted = 0;
    private int dragYstarted = 0;
    private int dragXcurrent = 0;
    private int dragYcurrent = 0;
    private boolean showBasePeak = false;
    private Vector listeners;
    
    public ChromatogramDisplayPanel (MZCollection collection) {
        this.collection = collection;
        msLevels = collection.getMSLevels();
        currentMSlevel = msLevels[0];
        spectra = collection.getSpectraForMSLevel(currentMSlevel);
        listeners = new Vector();
        
        setMinTime();
        setMaxTime();
        minIntensity = 0;
        setMaxIntensity();
        
        ClickNDragListener l = new ClickNDragListener(this);
        this.addMouseListener(l);
        this.addMouseMotionListener(l);
    }
    
    public void addSpectrumSelectionListener (SpectrumSelectionListener listener) {
    		listeners.add(listener);
    }
    
    public void useBasePeakPlot (boolean b) {
    		showBasePeak = b;
    		minIntensity = 0;
    		setMaxIntensity();
    		this.repaint();
    }

    public void setMSLevel (int i) {
		currentMSlevel = i;
		minIntensity = 0;
		spectra = collection.getSpectraForMSLevel(currentMSlevel);
		setMinTime();
		setMaxTime();
		setMaxIntensity();

        // Now let the listeners know
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((SpectrumSelectionListener)e.nextElement()).spectrumSelected(null);
		}
		this.repaint();
}

    
    private void setMaxIntensity() {
    		double tempMaxIntensity = 1;
    		for (int i=0;i<spectra.length;i++) {
    			if (spectra[i].getTime() < minTime || spectra[i].getTime() > maxTime) {
    				continue;
    			}
    			if (showBasePeak) {
    				if (spectra[i].getMaxIntensity() > tempMaxIntensity) {
    					tempMaxIntensity = spectra[i].getMaxIntensity();
    				}
    			}
    			else {
    				if (spectra[i].getTICValue() > tempMaxIntensity) {
    					tempMaxIntensity = spectra[i].getTICValue();
    				}
    			}
    		}
    		maxIntensity = tempMaxIntensity;
    }
    private void setMinIntensity() {
		double tempMinIntensity = 0;
		for (int i=0;i<spectra.length;i++) {
			if (spectra[i].getTime() > minTime || spectra[i].getTime() > maxTime) {
				continue;
			}
			if (showBasePeak) {
				if (spectra[i].getMaxIntensity() > tempMinIntensity) {
					tempMinIntensity = spectra[i].getMaxIntensity();
				}
			}
			else {
				if (spectra[i].getTICValue() > tempMinIntensity) {
					tempMinIntensity = spectra[i].getTICValue();
				}
			}
		}
		minIntensity = tempMinIntensity;
    }
    private void setMaxTime() {
    		float tempMaxTime = 0;
    		for (int i=0;i<spectra.length;i++) {
    			if (spectra[i].getTime() > tempMaxTime) {
    				tempMaxTime = spectra[i].getTime();
    			}
    		}
    		maxTime = tempMaxTime;
    }
    private void setMinTime() {
    		if (spectra.length >=1) {
    			float tempMinTime = spectra[0].getTime();
    			for (int i=1;i<spectra.length;i++) {
    				if (spectra[i].getTime() < tempMinTime) {
    					tempMinTime = spectra[i].getTime();
    				}
    			}
    			minTime = tempMinTime;
    		}
    }
    
    public void paint (Graphics g0) {
        super.paint(g0);
       
        Graphics2D g = (Graphics2D)g0;
               
        int width = this.getWidth();
        int height = this.getHeight();
        
        g.setColor(Color.WHITE);
        g.fillRect(0,0,width,height);
        
        g.setColor(Color.BLACK);
        g.setFont(new Font("sans-serif",Font.BOLD,12));

        
        if (showBasePeak) {
        		g.drawString("Base Peak Plot of msLevel "+currentMSlevel,5,15);
        }
        else {
        		g.drawString("TIC Plot of msLevel "+currentMSlevel,5,15);
        }
        
        // Draw the graph axes and legends
        g.setFont(new Font("sans-serif",Font.BOLD,12));
        g.drawLine(50,30,50,height-30);
        g.drawLine(50,height-30,width-10,height-30);
        String timeUnit = "";
        if (spectra.length > 0) {
        		timeUnit = "("+spectra[0].getTimeUnit()+")";
        }
        g.drawString("Time "+timeUnit,15+((width-60)/2),height-3);
        // Draw scale text
        
        // Time Scale (x-axis)
        DecimalFormat dfx;

        double timeDiff = maxTime-minTime;
        float timeBase = (float)timeDiff/5;

        int timeXscale = 0;
        float timeXincrease;
        while(true) {
            if (timeBase>10) {
                timeBase = timeBase/10;
                ++timeXscale;
            }
            else if (timeBase < 1) {
            		timeBase = timeBase*10;
            		--timeXscale;
            }
            else {
            		timeXincrease = (float)(Math.round(timeBase)*(float)(Math.pow(10,timeXscale)));
            		break;
            }
        }
        String formatString = "0";
        switch (timeXscale) {
        	case -1:
        	    formatString = "0.0";
        		break;
        	case -2:
        	    formatString = "0.00";
        		break;
        	case -3:
        	    formatString = "0.000";
        		break;
        	case -4:
        	    formatString = "0.0000";
        			break;
        }

        dfx = new DecimalFormat(formatString);


        float currentX = (float)((minTime+timeXincrease) - ((minTime)%timeXincrease));
        g.setFont(new Font("sans-serif",Font.PLAIN,10));
        while (currentX <= maxTime) {
            int xOffset = 50 + (int)((width-60)*((currentX-minTime)/(maxTime-minTime)));
            g.drawLine(xOffset,height-30,xOffset,height-25);
            g.drawString(""+dfx.format(currentX),xOffset-10,height-15);
            
            currentX += timeXincrease;
        }

        // Intensity Scale (y-axis)
        double intensityDiff = maxIntensity-minIntensity;
        double intensityBase = intensityDiff/5;
        DecimalFormat dfy = new DecimalFormat("0.#E0");
        int intensityYscale = 0;
        double intensityYincrease;
        while(true) {
            if (intensityBase>10) {
                intensityBase = intensityBase/10;
                ++intensityYscale;
            }
            else if (intensityBase < 1){
                intensityBase = intensityBase*10;
                --intensityYscale;
            }
            else {
                intensityYincrease = Math.round(intensityBase)*(Math.pow(10,intensityYscale));
                break;
            }
        }

        double currentY = (minIntensity+intensityYincrease) - ((minIntensity)%intensityYincrease);
        g.setFont(new Font("sans-serif",Font.PLAIN,10));
        while (currentY <= maxIntensity) {
            int yOffset = 30 + (int)((height-60)*((maxIntensity - currentY)/(maxIntensity-minIntensity)));
            g.drawLine(50,yOffset,45,yOffset);
            // Intensity values are BIG - so they get too long to fit into the
            // available space.
            g.drawString(dfy.format(currentY),5,yOffset+3);
            
            currentY += intensityYincrease;
        }
        
        
        // Shade the display to show zoom selections
        if (changingXZoom) {
            g.setColor(new Color(150,150,255));
            if (dragXcurrent >= dragXstarted) {
                g.fillRect(dragXstarted,30,dragXcurrent-dragXstarted,height-60);
            }
            else {
                g.fillRect(dragXcurrent,30,dragXstarted-dragXcurrent,height-60);                
            }
        }
        else if (changingYZoom) {
            g.setColor(new Color(150,150,255));
            if (dragYcurrent >= dragYstarted) {
                g.fillRect(50,dragYstarted,width-60,dragYcurrent-dragYstarted);
            }
            else {
                g.fillRect(50,dragYcurrent,width-60,dragYstarted-dragYcurrent);
            }
        }
                
        g.setColor(Color.BLACK);
        
        int lastXoffset=0;
        int lastYoffset=0;
        
        for (int i=0;i<spectra.length;i++) {

            if (spectra[i].getTime()<minTime || spectra[i].getTime()>maxTime) {
                continue;
            }
            
            if (showBasePeak) {
            		if (spectra[i].getMaxIntensity() < minIntensity) {
            			continue;
            		}
            }
            else {
            		if (spectra[i].getTICValue() < minIntensity) {
            			continue;
            		}            	
            }

            double drawnIntensity;
            
            if (showBasePeak) {
            		drawnIntensity = spectra[i].getMaxIntensity();
            }
            else {
            		drawnIntensity = spectra[i].getTICValue();
            }
            
            
            if (drawnIntensity > maxIntensity) {
            		drawnIntensity = maxIntensity;
            }
            
            int xOffset = (int)((width-60)*((spectra[i].getTime()-minTime)/(maxTime-minTime)));
            int yOffset = (int)((height-60)*((drawnIntensity-minIntensity)/(maxIntensity-minIntensity)));

            g.drawLine(50+lastXoffset,height-(30+lastYoffset),50+xOffset,height-(30+yOffset));
            
            lastXoffset = xOffset;
            lastYoffset = yOffset;
        }   
        //TODO: Find some way to draw the next point after the end is reached
    }
    
        
    private class ClickNDragListener implements MouseListener, MouseMotionListener {

        ChromatogramDisplayPanel d;
        
        public ClickNDragListener (ChromatogramDisplayPanel d) {
            this.d = d;
        }
        
        public void mouseClicked(MouseEvent me) {
            if (me.getClickCount()==2) {
                // A double click resets the zoom
                setMinTime();
                setMaxTime();
                minIntensity = 0;
                setMaxIntensity();
                changingXZoom = false;
                changingYZoom = false;
                d.repaint();
            }
            else if (me.getClickCount()==1) {
            		// We ignore clicks which are outside the main plot area
            		if (me.getX() < 50 || me.getX() > d.getWidth()-10 || me.getY() < 30 || me.getY()>d.getHeight()-30) {
            			return;
            		}
            		float selectedTime = minTime + ((maxTime-minTime)*((me.getX()-50)/((float)d.getWidth()-60)));
            		
            		float timeDiff = 0;
            		MZSpectrum closestSpectrum = null;
            		
            		for (int i=0;i<spectra.length;i++) {
            			float thisTimeDiff = selectedTime - spectra[i].getTime();
            			if (thisTimeDiff < 0) {
            				thisTimeDiff = 0-thisTimeDiff;
            			}
            			if (closestSpectrum == null) {
            				timeDiff = thisTimeDiff;
            				closestSpectrum = spectra[i];
            			}
            			else if (thisTimeDiff < timeDiff) {
            				timeDiff = thisTimeDiff;
            				closestSpectrum = spectra[i];
            			}
            		}
            		
            		// Now let the listeners know
            		Enumeration e = listeners.elements();
            		while (e.hasMoreElements()) {
            			((SpectrumSelectionListener)e.nextElement()).spectrumSelected(closestSpectrum);
            		}

            }
        }


        public void mousePressed(MouseEvent me) {
            dragXstarted = me.getX();
            dragXcurrent = me.getX();
            dragYstarted = me.getY();
            dragYcurrent = me.getY();
            
            if (dragXstarted <= 50 && dragYstarted >=30 && dragYstarted <= d.getHeight()-30) {
                changingYZoom = true;
            }
            else if (dragXstarted >= 50 && dragXstarted <= (d.getWidth()-10) && dragYstarted >= d.getHeight()-30) {
               changingXZoom = true;
            }

        }


        public void mouseReleased(MouseEvent arg0) {
            

            if (changingXZoom) {
                
                if (dragXstarted == dragXcurrent) {
                    changingXZoom = false;
                    d.repaint();
                    return;
                }
                int minX;
                int maxX;
                
                if (dragXstarted > dragXcurrent) {
                    minX = dragXcurrent;
                    maxX = dragXstarted;
                }
                else {
                    minX = dragXstarted;
                    maxX = dragXcurrent;                    
                }
                
                float tempMinTime = minTime + ((maxTime-minTime)*(((float)minX-50)/(d.getWidth()-60)));
                float tempMaxTime = minTime + ((maxTime-minTime)*(((float)maxX-50)/(d.getWidth()-60)));
                
                minTime=tempMinTime;
                maxTime=tempMaxTime;
                setMaxIntensity();
                
            }
            else if (changingYZoom) {
                if (dragYstarted == dragYcurrent) {
                    changingYZoom = false;
                    d.repaint();
                    return;
                }
                int minY;
                int maxY;
                
                if (dragYstarted > dragYcurrent) {
                    minY = dragYcurrent;
                    maxY = dragYstarted;
                }
                else {
                    minY = dragYstarted;
                    maxY = dragYcurrent;                    
                }
                
                double tempMinIntensity = maxIntensity - ((maxIntensity-minIntensity)*(((double)maxY-30)/(d.getHeight()-60)));
                double tempMaxIntensity = maxIntensity - ((maxIntensity-minIntensity)*(((double)minY-30)/(d.getHeight()-60)));
              
                minIntensity=tempMinIntensity;
                maxIntensity=tempMaxIntensity;
                

            }

            changingXZoom = false;
            changingYZoom = false;
            d.repaint();
        }


        public void mouseEntered(MouseEvent arg0) {
 
        }


        public void mouseExited(MouseEvent arg0) {
            changingXZoom = false;
            changingYZoom = false;            
        }


        public void mouseDragged(MouseEvent me) {
            dragXcurrent = me.getX();
            dragYcurrent = me.getY();
            
            if (dragXcurrent < 50) {
                dragXcurrent = 50;
            }
            if (dragXcurrent > d.getWidth()-10) {
                dragXcurrent = d.getWidth()-10;
            }
            if (dragYcurrent < 30) {
                dragYcurrent = 30;
            }
            if (dragYcurrent > d.getHeight()-30) {
                dragYcurrent = d.getHeight()-30;
            }

            d.repaint();
        }

        public void mouseMoved(MouseEvent arg0) {
 
        }

        
    }
}
